"""
MacroParse will be built atop the existing "mini" parsing and scanning tools.
It extends the language of attributed context-free grammars with additional features described on the wiki page.

The design for this module is still in flux, although most of the main ideas are laid out.
The concept is a single-file definition of both lexical and syntactic analysis embedded within MarkDown text.

Markdown as a container format is straightforward to analyze with available string operations or
simple applications of standard Python regex machinery. However, the miniscan/miniparse machinery is
extremely handy for recovering the syntactic structure of actual rules, so that's used here.

========================================

"""
import re, os, collections, typing
from ..support import foundation, compaction, interfaces
from ..parsing import automata
from ..scanning import regular, miniscan
from . import grammar


def compile_file(pathname, *, method, strict=False) -> dict:
	with(open(pathname)) as fh: document = fh.read()
	return compile_string(document, method=method).determinize().as_compact_form(filename=os.path.basename(pathname))

class TextBookForm:
	""" This provides the various views of the text-book form of scan and parse tables. """
	def __init__(self, *, dfa: regular.DFA, scan_actions:list, parse_table: automata.DragonBookTable):
		self.dfa = dfa
		self.scan_actions = scan_actions
		self.parse_table = parse_table
	def as_compact_form(self, *, filename):
		return {
			'description': 'MacroParse Automaton',
			'version': (0, 0, 1),
			'source': filename,
			'scanner': self.compact_scanner(),
			'parser': self.compact_parser(),
		}
	def compact_scanner(self):
		dfa = self.dfa
		if dfa is None: return
		return {
			'dfa': compaction.compress_scanner(initial=dfa.initial, matrix=dfa.states, final=dfa.final),
			'action': dict(zip(['message', 'parameter', 'trail', 'line_number'], zip(*self.scan_actions))),
			'alphabet': {'bounds': dfa.alphabet.bounds, 'classes': dfa.alphabet.classes,}
		}
	def compact_parser(self):
		table = self.parse_table
		if table is None: return
		symbol_index = {s: i for i, s in enumerate(table.terminals + table.nonterminals)}
		symbol_index[None] = None
		form = {
			'initial': table.initial,
			'action': compaction.compress_action_table(table.action_matrix, table.nonassoc_errors),
			'goto': compaction.compress_goto_table(table.goto_matrix),
			'terminals': table.terminals,
			'nonterminals': table.nonterminals,
			'breadcrumbs': [symbol_index[s] for s in table.breadcrumbs],
			'rule': encode_parse_rules(table.rule_table, table.constructors, table.rule_origin),
		}
		if table.splits: form['splits'] = table.splits
		return form
	def pretty_print(self):
		if self.dfa is not None:
			self.dfa.stats()
			self.dfa.display()
		if self.parse_table is not None:
			self.parse_table.display()
	def make_csv(self, pathstem):
		if self.dfa is not None:
			self.dfa.make_csv(pathstem)
		if self.parse_table is not None:
			self.parse_table.make_csv(pathstem)

class IntermediateForm(typing.NamedTuple):
	nfa: regular.NFA
	scan_actions: list
	hfa: automata.HFA
	parse_style:automata.ParsingStyle
	def determinize(self) -> TextBookForm:
		dfa = self.nfa.subset_construction().minimize_states().minimize_alphabet() if self.nfa.states else None
		return TextBookForm(dfa=dfa, scan_actions=self.scan_actions, parse_table=automata.tabulate(self.hfa, style=self.parse_style))
	def make_dot_file(self, path): self.hfa.make_dot_file(path)


def compile_string(document:str, *, method) -> IntermediateForm:
	""" This has the job of reading the specification and building the textbook-form tables. """
	# The approach is a sort of outside-in parse. The outermost layer concerns the overall markdown document format,
	# which is dealt with in the main body of this routine prior to determinizing and serializing everything.
	# Each major sub-language is line-oriented and interpreted with one of the following five subroutines:
	
	def handle_meta_exception(e: Exception):
		if isinstance(e, miniscan.PatternError):
			raise grammar.DefinitionError('At line %d: %s'%(line_number, e.args[0])) from None
		elif isinstance(e, interfaces.LanguageError):
			raise grammar.DefinitionError('At line %d: Malformed pattern.' % line_number) from None
		else: raise e

	def definitions():
		name, regex = current_line_text.split(None, 1)
		if name in env: raise grammar.DefinitionError('You cannot redefine named subexpression %r at line %d.'%(name, line_number))
		if not re.fullmatch(r'[A-Za-z][A-Za-z_]+', name): raise grammar.DefinitionError('Subexpression %r ought to obey the rule at line %d.'%(name, line_number))
		try: env[name] = miniscan.rex.parse(miniscan.META.scan(regex, env=env), language='Regular')
		except Exception as e: handle_meta_exception(e)
		assert isinstance(env[name], regular.Regular), "This would be a bug."
	
	def conditions():
		"""
		The first token will be a condition name. Thereafter, maybe an arrow and one or more included groups.
		Pattern groups named on the LEFT hand side are real start conditions, accessible in the final scanner.
		Those which appear only on the right hand side are "virtual", usable only by inclusion.
		At some point it might be nice to add validation that these are all used correctly...
		"""
		name, includes = error_help.parse(current_line_text, line_number, "condition")
		if name in condition_definitions: error_help.gripe('Re-declared scan-condition %r; this is unexpected.'%name)
		condition_definitions[name] = includes
	
	def patterns():
		# This could be done better: a nice exercise might be to enhance the present regex parser to also
		# grok actual scanner rules as an alternate language start-symbol; such could eliminate some of
		# this contemptible string hackery and thereby enable things like embedded spaces where they make sense.
		# Such would also involve hacking the metascanner bootstrap code to track paren depth and recognize
		# the other tokens that can appear.
		def note_pattern(pattern):
			# Now patterns that share a trail length can also share a rule ID number.
			try: bol, expression, trail = miniscan.analyze_pattern(pattern, env)
			except Exception as e: handle_meta_exception(e)
			else: pending_patterns[trail].append((bol, expression))
		if current_line_text.endswith('|'):
			pattern = current_line_text[:-1].strip()
			if re.search(r'\s', pattern): raise grammar.DefinitionError('Unable to analyze pattern/same-as-next structure at line %d.')
			note_pattern(pattern)
			return
		m = re.fullmatch(r'(.*?)\s*:([A-Za-z][A-Za-z_]*)(?:\s+([A-Za-z_]+))?(?:\s+:(0|[1-9][0-9]*))?', current_line_text)
		if not m: raise grammar.DefinitionError('Unable to analyze overall pattern/action/parameter/(rank) structure at line %d.'%line_number)
		pattern, action, parameter, rank_string = m.groups()
		rank = int(rank_string) if rank_string else 0
		note_pattern(pattern)
		for trail, list_of_patterns in pending_patterns.items():
			rule_id = foundation.allocate(scan_actions, (action, parameter, trail, line_number))
			for bol, expression in list_of_patterns:
				src = nfa.new_node(rank)
				dst = nfa.new_node(rank)
				for q,b in zip(nfa.condition(current_pattern_group), bol):
					if b: nfa.link_epsilon(q, src)
				nfa.final[dst] = rule_id
				expression.encode(src, dst, nfa, rank)
		pending_patterns.clear()
		pass
	
	def precedence(): ebnf.read_precedence_line(current_line_text, line_number)
	
	def productions(): ebnf.read_production_line(current_line_text, line_number)
	
	def decide_section():
		# Looks at a header line to see which parsing mode/section to shift into based on a leading keyword,
		# and also performs any clerical duties associated with said shift.
		tokens = ''.join([c if c.isalnum() or c=='_' else ' ' for c in current_line_text]).split()
		if not tokens: return None
		head = tokens[0].lower()
		if head == 'definitions': return definitions
		if head == 'conditions':
			# The way to handle it is to set up epsilon-connections between the conditions
			# as specified in the source definition, and then delete "virtual" conditions
			# from nfa.initial before performing the subset construction. If no "virtual"
			# conditions are determined, then there's nothing to delete, and all groups get presented.
			return conditions
		if head == 'patterns':
			nonlocal current_pattern_group
			current_pattern_group = tokens[1] if len(tokens)>1 else interfaces.DEFAULT_INITIAL_CONDITION
			return patterns
		if head == 'precedence': return precedence
		if head == 'productions':
			ebnf.current_head = None
			for t in tokens[1:]:
				if t not in ebnf.plain_cfg.start:
					ebnf.plain_cfg.start.append(t)
			return productions
		return None

	# The context-free portion of the definition:
	error_help = grammar.ErrorHelper()
	ebnf = grammar.EBNF_Definition(error_help)
	
	# The regular (finite-state) portion of the definition:
	env = miniscan.PRELOAD['ASCII'].copy()
	nfa = regular.NFA()
	pending_patterns = collections.defaultdict(list) # Those awaiting an application of the `|` action...
	scan_actions = [] # That of a regular-language rule entry is <message, parameter, trail, line_number>
	current_pattern_group = None
	condition_definitions = {}
	
	def tie_conditions():
		declared = set(condition_definitions.keys())
		declared.update(*condition_definitions.values())
		forgot_to_define = declared - set(nfa.initial.keys())
		if forgot_to_define: raise grammar.DefinitionError('These pattern groups were declared in the conditions block but never defined:\n'+repr(forgot_to_define))
		forgot_to_declare = set(nfa.initial.keys()) - declared
		if forgot_to_declare: raise grammar.DefinitionError('These pattern groups appear, but are not declared in the conditions block:\n'+repr(forgot_to_declare))
		# TODO: Check for no cycles in the inclusion graph...
		virtual_groups = declared - set(condition_definitions.keys())
		for name, includes in condition_definitions.items():
			for i in includes:
				nfa.link_condition(name, i)
		for name in virtual_groups:
			del nfa.initial[name]

	# Here begins the outermost layer of grammar definition parsing, which is to comprehend the
	# structure of a supplied mark-down document just enough to extract headers and code-blocks.
	section, in_code, line_number = None, False, 0
	for current_line_text in document.splitlines(keepends=False):
		line_number += 1
		if in_code:
			current_line_text = current_line_text.strip()
			if '```' in current_line_text:
				in_code = False
				if pending_patterns: raise grammar.DefinitionError("Consecutive group of patterns lacks a scanner action before end of code block at line %d."%line_number)
			elif current_line_text and section: section()
		elif current_line_text.startswith('#'): section = decide_section()
		elif current_line_text.strip().startswith('```'): in_code = True
		else: continue
	if in_code and section: raise grammar.DefinitionError("A code block fails to terminate before the end of the document.")
	
	# Compose the control tables. (Compaction is elsewhere. Serialization will be straight JSON via standard library.)
	if condition_definitions: tie_conditions()
	hfa = automata.PARSE_TABLE_METHODS[method](ebnf.sugarless_form())
	if ebnf.nondeterministic_symbols:
		style = automata.GeneralizedStyle(len(hfa.graph), ebnf.nondeterministic_symbols)
	else:
		style = automata.DeterministicStyle(False)
	return IntermediateForm(nfa=nfa, scan_actions=scan_actions, hfa=hfa, parse_style=style,)


def encode_parse_rules(rules:list, constructors:list, origins:list) -> dict:
	assert isinstance(rules, list), type(rules)
	return {'rules': rules, 'line_number': origins, 'constructor': constructors, }

def main():
	import sys, argparse, json
	parser = argparse.ArgumentParser(
		prog='py -m boozetools.macroparse.compiler',
		description='Compile a macroparse grammar/scanner definition from a markdown document into a set of parsing and scanning tables in JSON format.',
	)
	parser.add_argument('source_path', help='path to input file')
	parser.add_argument('-f', '--force', action='store_true', dest='force', help='allow to write over existing file')
	parser.add_argument('-o', '--output', help='path to output file')
	parser.add_argument('-i', '--indent', help='indent the JSON output for easier reading.', action='store_const', dest='indent', const=2, default=None)
	parser.add_argument('--pretty', action='store_true', help='Display uncompressed tables in attractive grid format on STDOUT.')
	parser.add_argument('--csv', action='store_true', help='Generate CSV versions of uncompressed tables, suitable for inspection.')
	parser.add_argument('--dev', action='store_true', help='Operate in "development mode" -- which changes from time to time.')
	parser.add_argument('--dot', action='store_true', help="Create a .dot file for visualizing the parser via the Graphviz package.")
	parser.add_argument('-m', '--method', choices=automata.PARSE_TABLE_METHODS, default='LR1', type=str.upper, help="Which parser table construction method to use.")
	parser.add_argument('-v', '--verbose', action='store_true', help="Squawk, mainly about the table compression stats.")
	if len(sys.argv) < 2: exit(parser.print_help())
	args = parser.parse_args()
	if args.verbose: compaction.VERBOSE = True
	stem, extension = os.path.splitext(args.source_path)
	target_path = args.output or stem+'.automaton'
	if os.path.exists(target_path) and not args.force:
		print('Target file already exists and --force command-line argument was not given.', file=sys.stderr)
		exit(1)
	with(open(args.source_path)) as fh:document = fh.read()
	try:
		intermediate_form = compile_string(document, method=args.method)
		if args.dot: intermediate_form.make_dot_file(target_path+'.dot')
		textbook_form = intermediate_form.determinize()
	except grammar.DefinitionError as e:
		print(e.args[0], file=sys.stderr)
		exit(1)
	else:
		compact = textbook_form.as_compact_form(filename=os.path.basename(args.source_path))
		if args.pretty: textbook_form.pretty_print()
		if args.csv: textbook_form.make_csv(target_path)
		if args.dev:
			compact_goto = compact['parser']['goto']
			for listname in ['row_index', 'col_index', 'quotient']:
				print(listname+':', compact_goto[listname], len(compact_goto[listname]))
			print('mark:', compact_goto['mark'], 'residue:', len(compact_goto['quotient'])-compact_goto['mark'])
		json.dump(compact, open(target_path, 'w'), separators = (',', ':'), sort_keys = False, indent = args.indent)
		print('Wrote automaton in JSON format to:')
		print('\t'+target_path)

if __name__ == '__main__': main()



"""
Provide a mechanism to read the compacted scanning and parsing tables,
prepared using the algorithms in .compaction.py. If you were to build
a runtime system for another host language, you'd have to port these.
"""

from typing import Callable
from . import interfaces
from ..scanning import charclass

def displacement_function(otherwise:callable, *, offset, check, value) -> callable:
	"""
	(make a) Reader for a perfect-hash such as built at compaction.encode_displacement_function.
	If the queried cell is absent, the `otherwise` functional-argument provides the back-up plan.
	I'll admit that form is a bit unusual in Python, but it DOES encapsulate the interpretation.
	"""
	size = len(check)
	def fn(row_id:int, column:int) -> int:
		index = offset[row_id] + column
		if 0 <= index < size and check[index] == row_id:
			return value[index]
		else:
			return otherwise(row_id, column)
	return fn

def boolean_field(row_class, col_class, offset, check) -> Callable[[int, int], bool]:
	"""
	Make a reader for a boolean field. It's used in a couple places.
	They're compressed by equivalence-classification along both dimensions.
	"""
	size = len(check)
	def fn(row:int, col:int) -> bool:
		try: rc, cc = row_class[row], col_class[col]
		except IndexError: return False # The bogus-token is out of range.
		index = offset[rc] + cc
		return 0 <= index < size and check[index] == rc
	return fn

def scanner_delta_function(*, exceptions:dict, bg:dict) -> callable:
	"""
	This function came out of some observations about the statistics of scanner delta functions.
	The best way to understand this function is by reference to the following URL:
	https://github.com/kjosib/booze-tools/issues/8
	
	The basic idea is to first check a table of "unusual" entries in the scanner's transition matrix.
	If that comes up empty, then the "background" table is essentially a bitmap combined with a pair
	of most-common entries per given row. The bitmap compresses very well by means of row- and
	column-equivalence classes; only the 1's (being the less common) need be actually represented.
	
	The very most common entry in a row is usually the error transition (-1), so instead of a
	complete list per state, only the exceptions to this rule are listed.
	"""
	zeros = dict(zip(*bg['zero']))
	broad_pattern = boolean_field(bg['row_class'], bg['col_class'], bg['offset'], bg['check'])
	def general_population(state_id:int, symbol_id:int) -> int:
		if broad_pattern(state_id, symbol_id): return bg['one'][state_id]
		else: return zeros.get(state_id, -1)
	return displacement_function(general_population, **exceptions)

def parser_action_function(*, reduce, fallback, edits) -> callable:
	background = parser_reduce_function(**reduce)
	offset, check, value = edits['offset'], edits['check'], edits['value']
	size = len(check)
	def chase(row, col):
		probe = row
		while probe >= 0:
			index = offset[probe] + col
			if 0 <= index < size and check[index] == probe: return value[index]
			probe = fallback[probe]
		return background(row, col)
	absent = len(edits['check'])
	interactive = [r if o == absent and f==-1 else 0 for r,o,f in zip(reduce['d_reduce'], edits['offset'], fallback)]
	return chase, interactive.__getitem__

def parser_goto_function(*, row_index, col_index, quotient, mark ) -> callable:
	def probe(state_id:int, nonterminal_id:int):
		r, c = row_index[state_id], col_index[nonterminal_id]
		dominant = min(r, c)
		return quotient[dominant] if dominant < mark else quotient[r+c-mark]
	return probe

def parser_reduce_function(*, d_reduce, row_class, col_class, offset, check):
	predicate = boolean_field(row_class, col_class, offset, check)
	def fn(row, col):
		dr = d_reduce[row] # Consulting the predicate costs a few cycles.
		return dr if dr and predicate(row, col) else 0 # Do it only at need.
	return fn

class CompactDFA(interfaces.FiniteAutomaton):
	"""
	This implements the FiniteAutomaton interface (for use with the generic scanner algorithm)
	by reference to a set of scanner tables that have been built using the MacroParse machinery.
	It's not the whole story; BoundScanRules (defined below) are involved in binding the
	action specifications to a specific context object.
	"""
	def __init__(self, *, dfa:dict, alphabet:dict):
		self.classifier = charclass.MetaClassifier(**alphabet)
		self.delta = scanner_delta_function(**dfa['delta'])
		self.initial = dfa['initial']
		self.final = dict(zip(dfa['final'], dfa['rule']))
	
	def jam_state(self): return -1
	def get_condition(self, condition_name) -> tuple:
		try: return self.initial[condition_name]
		except KeyError: raise KeyError(condition_name, set(self.initial.keys()))
	def get_state_rule_id(self, state_id: int) -> int: return self.final.get(state_id)

	def get_next_state(self, current_state: int, codepoint: int) -> int:
		return self.delta(current_state, self.classifier.classify(codepoint))
	
class CompactHandleFindingAutomaton(interfaces.ParseTable):
	"""
	This implements the ParseTable interface (for use with the generic parse algorithm)
	by reference to a set of parser tables that have been built using the MacroParse machinery.
	It's not the whole story: something needs to provide reduction bindings.
	"""
	def __init__(self, parser:dict):
		self.get_action, self.interactive_step = parser_action_function(**parser['action'])
		self.get_goto = parser_goto_function(**parser['goto'])
		self.terminals = parser['terminals']
		self.__translation = {symbol:i for i,symbol in enumerate(self.terminals)}
		self.nonterminals = parser['nonterminals']
		self.initial = parser['initial']
		self.breadcrumbs = parser['breadcrumbs']
		self.__rule = parser['rule']['rules']
		self.message_catalog = parser['rule']['constructor']
		if 'splits' in parser:
			self.get_split_offset = parser['action']['reduce']['d_reduce'].__len__ # Gets the number of states.
			self.get_split = parser['splits'].__getitem__
		
	def get_translation(self, symbol) -> int:
		try: return self.__translation[symbol]
		except KeyError: return len(self.terminals) # Guaranteed to trigger error-processing.
	
	def get_action(self, state_id: int, terminal_id) -> int: assert False, 'See the constructor.'
	def get_goto(self, state_id: int, nonterminal_id) -> int: assert False, 'See the constructor.'
	def get_rule(self, rule_id: int) -> tuple: return self.__rule[rule_id]
	def get_constructor(self, constructor_id) -> object: return self.message_catalog[constructor_id]
	def get_initial(self, language) -> int: return 0 if language is None else self.initial[language]
	def get_breadcrumb(self, state_id: int) -> str:
		bcid = self.breadcrumbs[state_id]
		if bcid is None: return ''
		if bcid < len(self.terminals): return self.terminals[bcid]
		else: return self.nonterminals[bcid-len(self.terminals)]
	def interactive_step(self, state_id: int) -> int: assert False, 'See the constructor.'
	def get_split_offset(self) -> int: assert False, 'See the constructor.'
	def get_split(self, split_id: int) -> list: assert False, 'See the constructor.'

def _expand_rules(raw):
	"""
	Implement a proposed interpretation of a tight rules table.
	There's a related idea that might work in a C-like language where the rule IDs
	in the ACTION table are pre-translated to the offset (from the end) of their
	position in the rule data table. Then you just need a sentinel 0 at the end.
	"""
	result = []
	i = 0
	while i < len(raw):
		nonterminal_id, rhs_length, constructor_id = raw[i:i+3]
		view = []
		i += 3
		while i < len(raw) and raw[i] < 0:
			view.append(raw[i])
			i += 1
		result.append((nonterminal_id, rhs_length, constructor_id, tuple(view)))
	return result



""" Bits and bobs in support of visualizing data structures. """
import csv

DOT = '\u25cf'


def print_grid(grid):
	lens = list(map(len, grid))
	assert len(set(lens)) == 1, lens
	grid = [[str(cell) for cell in row] for row in grid]
	width = [max(map(len, column)) for column in zip(*grid)]
	horizontal = '\u2500'
	vertical = ' \u2502 '
	upper = horizontal + '\u252c' + horizontal
	inner = horizontal + '\u253c' + horizontal
	lower = horizontal + '\u2534' + horizontal
	# horizontal = '-'
	# vertical = ' | '
	# upper = horizontal + '+' + horizontal
	# inner = horizontal + '+' + horizontal
	# lower = horizontal + '+' + horizontal
	segments = [horizontal*w for w in width]
	divider = inner.join(segments)
	print(upper.join(segments))
	for r, row in enumerate(grid):
		if r %5 == 1: print(divider)
		print(vertical.join(s.rjust(w,' ') for s,w in zip(row, width)))
	print(lower.join(segments))

def write_csv_grid(path, grid):
	with open(path, 'w', newline="") as fh:
		csv.writer(fh).writerows(grid)



"""
Things Prone to Failure:

1. People.
2. Machines.

This module is all about easing over the process to display where things go wrong.
A tool-developer should have easy access to sophisticated error-displays.
Such should be made both likely and comfortable independent of other tools.

Error reporting is actually a topic of some complexity. Fortunately, the common cases
are reasonably straightforward and for the uncommon ones there is YAGNI (yet).

If you can localize where an error came from, you'd generally like to include some
context in the report. The usual strategy is to show the offending line, ideally
with a specific portion highlighted somehow. If you're dealing with a text console
(as many tools do) then the `illustration` function helps: Given a single line of
text and a few parameters, it makes a suitable picture.

I prefer to keep line-breaking separate from scanning and parsing, so in those areas
a simple integer position is as much location data as you get. What's a decent means
to convert that to a line and column number? Or to slice the corresponding line
from a larger text string? The SourceText handles that. It also provides a nice
`complain(...)` method, which formats a decent-looking error message to STDOUT.

There is just one complication:

Line breaks are a funny thing. Unix calls for \n. Apple prior to OSx called for \r.
CP/M and its derivatives like Windows call for \r\n, really a printer control sequence.
OS/390 (an IBM mainframe thing) tends toward character 0x85 (a high-ascii control code)
whenever it isn't thinking in EBCDIC, but who in their right mind uses EBCDIC anymore?
The Unicode line-breaking algorithm calls for no less than ELEVEN ways to break a line!

The applications I've tested treat the Unix, Apple, and DOS conventions as line-breaks
and mostly ignore the other options defined in the Unicode standard, so that's the
default behavior of the SourceText. But you can supply a mode parameter to specify
different line-ending conventions. The options are given symbolically as keys in the
LINEBREAK_MODE dictionary.

So how exactly SHOULD you delimit lines? The answer, my friend, is blowin' in the wind....
"""

import bisect, re, sys

LINEBREAK_MODE = {
	'normal': re.compile(r'\r\n?|\n'),
	'unicode': re.compile(r'\r\n|[\x0a-\x0d\x1c-\x1e\u0085\u2028\u2029]]'),
	'unix': re.compile(r'\n'),
	'apple': re.compile(r'\r'),
	'dos': re.compile(r'\r\n'),
}

def illustration(single_line:str, start:int, width:int=0, *, prefix='') -> str:
	""" Builds up a picture of where something appears in a line of text. Useful for polite error messages. """
	blanks = ''.join(c if c == '\t' else ' ' for c in prefix + single_line[:start])
	underline = '^'*(width or 1)+'-- near here'
	return prefix + single_line.rstrip() + '\n' + blanks + underline

class SourceText:
	""" A wrapper for e.g. a program source text which can print a half-respectable error message with context. """
	def __init__(self, content:str, line_breaks='normal', filename:str=None):
		self.content = content
		self.filename = filename
		self.line_breaks = line_breaks
	
	def __make_bounds(self):
		""" Lazily only find line breaks if it turns out to be necessary for a particular text. """
		if not hasattr(self, '__bound'):
			inside = [m.end() for m in LINEBREAK_MODE[self.line_breaks].finditer(self.content)]
			self.__bounds = [0] + inside + [len(self.content)]

	def find_row_col(self, index:int):
		""" Zero-Based! This right here is the main point of the class... """
		self.__make_bounds()
		row = bisect.bisect_right(self.__bounds, index, hi=len(self.__bounds) - 1) - 1
		col = index - self.__bounds[row]
		return row, col
	
	def complain(self, index:int, width=1, *, message:str=""):
		""" Present an informative error message with an illustration of context. """
		row, col = self.find_row_col(index)
		line = self.content[self.__bounds[row]:self.__bounds[row + 1]]
		if self.filename is None: print("At line %d, column %d:"%(row+1, col+1), message, file=sys.stderr)
		else: print("At line %d, column %d, in %s:"%(row+1, col+1, self.filename), message, file=sys.stderr)
		print(illustration(line, col, width, prefix=' >>> '), file=sys.stderr)



""" Hook regular expression patterns up to method calls on a scanner object. """
from boozetools.support.interfaces import Scanner

from ..support import interfaces
from ..parsing import miniparse
from . import regular, charset, recognition

PRELOAD = {'ASCII': {k: regular.CharClass(cls) for k, cls in charset.POSIX.items()}}

class PatternError(Exception): pass
class BadReferenceError(PatternError):
	"""
	Raised if a pattern refers to a subexpression name not defined in the context of the analysis.
	args are the offending reference and the span within the pattern (offset/length pair) where it appears.
	"""
class TrailingContextError(PatternError):
	"""
	Raised if pattern has both variable-sized stem and variable-sized trailing context.
	This is not supported at this time.
	"""

class Definition(interfaces.ScanRules):
	def __init__(self, *, minimize=True, mode='ASCII'):
		self.__actions = []
		self.__trails = []
		self.__subexpressions = PRELOAD[mode].copy()
		self.__dfa: regular.DFA = None
		self.__nfa = regular.NFA()
		self.__minimize = minimize
		self.__awaiting_action = False
	
	def get_trailing_context(self, rule_id: int): return self.__trails[rule_id]
	
	def get_dfa(self) -> interfaces.FiniteAutomaton:
		if self.__dfa is None:
			if self.__awaiting_action: raise AssertionError('You forgot to provide the action for the final pattern!')
			self.__dfa = self.__nfa.subset_construction()
			if self.__minimize: self.__dfa = self.__dfa.minimize_states().minimize_alphabet()
		return self.__dfa
	
	def scan(self, text, *, start=None, env=None, on_error:interfaces.ScanErrorListener = None):
		if on_error is None: on_error = interfaces.ScanErrorListener()
		scanner = recognition.IterableScanner(text=text, automaton=self.get_dfa(), rules=self, start=start, on_error=on_error)
		scanner.env = env
		return scanner
		
	def install_subexpression(self, name:str, expression: regular.Regular):
		assert isinstance(name, str) and name not in self.__subexpressions and len(name) > 1
		assert isinstance(expression, regular.Regular)
		self.__subexpressions[name] = expression
	
	def install_rule(self, *, action:callable, expression: regular.Regular, bol=(True, True), condition: (str, list, tuple)=None, trail:int=None, rank:int=0) -> int:
		rule_id = len(self.__actions)
		self.__actions.append(action)
		self.__trails.append(trail)
		src = self.__nfa.new_node(rank)
		dst = self.__nfa.new_node(rank)
		if condition is None or isinstance(condition, str): condition = [condition]
		else: assert isinstance(condition, (tuple, list)), type(condition)
		for C in condition:
			for q, b in zip(self.__nfa.condition(C), bol):
				if b: self.__nfa.link_epsilon(q,src)
		self.__nfa.final[dst] = rule_id
		expression.encode(src, dst, self.__nfa, rank)
		return rule_id
	
	def let(self, name:str, pattern:str):
		self.install_subexpression(name, rex.parse(META.scan(pattern, env=self.__subexpressions), language='Regular'))
	
	def token(self, kind:str, pattern:str, *, rank=0, condition=None):
		""" This says every member of the pattern has token kind=kind and semantic=matched text. """
		@self.on(pattern, rank=rank, condition=condition)
		def action(yy:Scanner): yy.token(kind, yy.matched_text())
	
	def token_map(self, kind:str, pattern:str, fn:callable, *, rank=0, condition=None):
		""" Every member of the pattern has token kind=kind and semantic=fn(matched text). """
		@self.on(pattern, rank=rank, condition=condition)
		def action(yy:Scanner): yy.token(kind, fn(yy.matched_text()))

	def ignore(self, pattern:str, *, rank=0, condition=None):
		""" Tell Scanner to ignore what matches the pattern. """
		@self.on(pattern, rank=rank, condition=condition)
		def action(yy:Scanner): pass

	def on(self, pattern:str, *, condition=None, rank=0):
		"""
		For instance:
		scanner_definition.on('r#[^\n]+')(None) # Ignore comments
		@scanner_definition.on(r'[A-Za-z_]+')
		def word(yy): yy.token('word', scanner.matched_text())
		"""
		if self.__awaiting_action: raise AssertionError('You forgot to provide the action for the previous pattern!')
		self.__awaiting_action = True
		bol, expression, trail = analyze_pattern(pattern, self.__subexpressions)
		def decorator(fn):
			assert self.__awaiting_action
			self.__awaiting_action = False
			assert callable(fn)
			self.install_rule(action=fn, expression=expression, bol=bol, condition=condition, trail=trail, rank=rank)
			return fn
		return decorator
	def condition(self, *condition): return ConditionContext(self, *condition)
	
	def invoke(self, yy: interfaces.Scanner, rule_id:int):
		action = self.__actions[rule_id]
		if callable(action): action(yy)
		else: assert action is None
	


def analyze_pattern(pattern:str, env):
	scanner = META.scan(pattern, env=env)
	try: bol, expression, trailing_context = rex.parse(scanner)
	except BadReferenceError as e:
		pos,size = scanner.current_span()
		raise BadReferenceError("column %d: %r %s"%(pos+1, pattern[pos:pos+size], e.args[0]))
	assert isinstance(expression, regular.Regular)
	if not trailing_context: trail = None
	else:
		assert isinstance(trailing_context, regular.Regular), trailing_context
		stem, trail = expression.length(), trailing_context.length()
		expression = regular.Sequence(expression, trailing_context)
		if trail: trail = -trail
		elif stem: trail = stem
		else: raise TrailingContextError('Variable stem and variable trailing context in the same pattern are not presently supported.')
	return bol, expression, trail

class ConditionContext:
	""" I'd like to be able to use Python's context manager protocol to simplify writing definitions of scan conditions. """
	def __init__(self, definition:Definition, *condition):
		self.__definition = definition
		self.__condition = condition
	def __enter__(self): return self
	def __exit__(self, exc_type, exc_val, exc_tb): pass
	def on(self, pattern, *, rank=0): return self.__definition.on(pattern, rank=rank, condition=self.__condition)
	def install_rule(self, **kwargs): return self.__definition.install_rule(condition=self.__condition, **kwargs)
	def token(self, kind:str, pattern:str, *, rank=0): self.__definition.token(kind, pattern, rank=rank, condition=self.__condition)
	def token_map(self, kind:str, pattern:str, fn:callable, *, rank=0): self.__definition.token_map(kind, pattern, fn, rank=rank, condition=self.__condition)
	def ignore(self, pattern:str, *, rank=0): self.__definition.ignore(pattern, rank=rank, condition=self.__condition)

#########################
# A pattern parser is easy to build using the miniparse module:

rex = miniparse.MiniParse('Pattern', 'Regular')
rex.rule('Pattern', 'BOL Regular Trail')()
rex.rule('Pattern', 'BOL $')(lambda bol, eof:(bol, eof, None)) # To specify end-of-file rules without introducing yet another metacharacter
rex.rule('BOL', '')(lambda : (True, True))
rex.rule('BOL', '^')(lambda x: (False, True))
rex.rule('BOL', '^^')(lambda x: (True, False))
rex.rule('Trail', '')()
rex.rule('Trail', '$')()
rex.rule('Trail', '/ .Regular')()
rex.rule('Trail', '/ .Regular .$')(regular.Sequence)
rex.rule('Regular', 'Sequence')()
rex.rule('Regular', '.Regular | .Sequence')(regular.Alternation)
rex.rule('Sequence', 'Term')()
rex.rule('Sequence', 'Sequence Term')(regular.Sequence)
rex.rule('Term', 'Atom')()
rex.rule('Term', '.Atom ?')(regular.Hook)
rex.rule('Term', '.Atom *')(regular.Star)
rex.rule('Term', '.Atom +')(regular.Plus)
rex.rule('Term', '.Atom { .number }')(lambda a, n: regular.Counted(a, n, n))
rex.rule('Term', '.Atom { .number , }')(lambda a, n: regular.Counted(a, n, None))
rex.rule('Term', '.Atom { , .number }')(lambda a, n: regular.Counted(a, 0, n))
rex.rule('Term', '.Atom { .number , .number }')(lambda a, m, n: regular.Counted(a, m, n))
rex.rule('Atom', '( .Regular )')()
rex.rule('Atom', 'c')(lambda c: regular.CharClass([c, c + 1]))
rex.rule('Atom', 'reference')(None)
rex.rule('Atom', '[ .Class ]')(regular.CharClass)
rex.rule('Class', 'Conjunct')()
rex.rule('Class', '.Class && .Conjunct')(charset.intersect)
rex.rule('Conjunct', 'Members')()
rex.rule('Conjunct', '^ .Members')(charset.complement)
rex.rule('Members', 'Item')()
rex.rule('Members', 'Members Item')(charset.union)
rex.rule('Item', 'c')(charset.singleton)
rex.rule('Item', '.c - .c')(charset.range_class)
rex.rule('Item', 'short')(lambda c:c.cls)
@rex.rule('Item', 'reference')
def classref(subex: regular.Regular):
	if isinstance(subex, regular.CharClass): return subex.cls
	else: raise BadReferenceError('Reference is not to a character class, and so cannot be used within one.')

META = Definition()
def _BEGIN_():
	""" This is the bootstrapping routine: it builds the preload and the meta-scanner. """
	def seq(head, *tail):
		for t in tail: head = regular.Sequence(head, t)
		return head
	def txt(s):return seq(*(regular.CharClass(charset.singleton(ord(_))) for _ in s))
	
	def _metatoken(yy): yy.token(yy.matched_text(), None)
	def _and_then(condition):
		def fn(scanner):
			_metatoken(scanner)
			scanner.enter(condition)
		return fn
	def _instead(condition):
		def fn(scanner):
			scanner.less(0)
			scanner.enter(condition)
		return fn
	def _bracket_reference(yy):
		name = yy.matched_text()[1:-1]
		try: yy.token('reference', yy.env[name])
		except KeyError: raise BadReferenceError("Undefined sub-pattern reference")
	def _shorthand_reference(yy): yy.token('reference', yy.env[yy.matched_text()[1]])
	def _dot_reference(yy): yy.token('reference', yy.env['DOT'])
	def _hex_escape(yy): yy.token('c', int(yy.matched_text()[2:], 16))
	def _control(yy): yy.token('c', 31 & ord(yy.matched_text()[2:]))
	def _arbitrary_character(yy): yy.token('c', ord(yy.matched_text()))
	def _class_initial_close_bracket(yy):
		yy.enter('[')
		_arbitrary_character(yy)
	def _arbitrary_escape(yy): yy.token('c', ord(yy.matched_text()[1:]))
	def _number(yy): yy.token('number', int(yy.matched_text()))
	def _dollar(charclass):
		def fn(yy:Scanner): yy.token('$', charclass)
		return fn
	
	dot = PRELOAD['ASCII']['DOT'] = regular.CharClass([0, 10, 14])
	for codepoint, char in [(0, '0'), (27, 'e'), *enumerate('abtnvfr', 7)]: PRELOAD['ASCII'][char] = regular.CharClass(
		charset.singleton(codepoint))
	for codepoint, mnemonic in enumerate('NUL SOH STX ETX EOT ENQ ACK BEL BS TAB LF VT FF CR SO SI DLE DC1 DC2 DC3 DC4 NAK SYN ETB CAN EM SUB ESC FS GS RS US SP'.split()):
		PRELOAD['ASCII'][mnemonic] = regular.CharClass(charset.singleton(codepoint))
	PRELOAD['ASCII']['DEL'] = regular.CharClass(charset.singleton(127))
	PRELOAD['ASCII']['ANY'] = regular.CharClass(charset.UNIVERSAL)
	PRELOAD['ASCII']['vertical'] = regular.CharClass([10, 14])
	PRELOAD['ASCII']['horizontal'] = regular.CharClass([8, 10, 32, 33])
	for shorthand, longhand in [
		('d', 'digit'),
		('l', 'alpha'),
		('w', 'word'),
		('s', 'space'),
		('h', 'horizontal'),
	]:
		PRELOAD['ASCII'][shorthand] = PRELOAD['ASCII'][longhand]
		PRELOAD['ASCII'][shorthand.upper()] = regular.CharClass(
			charset.subtract(dot.cls, PRELOAD['ASCII'][longhand].cls))
	def ref(x): return PRELOAD['ASCII'][x]
	
	eof_charclass = regular.CharClass(charset.EOF)
	dollar_charclass = regular.CharClass(charset.union(charset.EOF, PRELOAD['ASCII']['vertical'].cls))
	META.install_rule(expression=txt('^'), action=_metatoken, bol=(False, True))
	META.install_rule(expression=txt('^^'), action=_metatoken, bol=(False, True))
	# This next rule says that a dollar-sign at the end of a pattern supplies a regex matching EITHER end-of-line OR end-of-file
	META.install_rule(expression=seq(txt('$'), eof_charclass), trail=-1, action=_dollar(dollar_charclass))
	# So I build another similar one for end-of-file rules: the string '<<EOF>>' appearing at the end of the pattern.
	# And by the way, the rex grammar (above) correctly directs such things to the trailing-context fork of the parse.
	META.install_rule(expression=seq(txt('<<EOF>>'), eof_charclass), trail=-1, action=_dollar(eof_charclass))
	for c in '(|)?*+/': META.install_rule(expression=txt(c), action=_metatoken)
	META.install_rule(expression=txt('.'), action=_dot_reference)
	META.install_rule(expression=txt('{'), action=_and_then('{'))
	META.install_rule(expression=txt('['), action=_and_then('[.'))
	META.install_rule(expression=txt('[^'), trail=-1, action=_and_then('^'))
	META.install_rule(expression=txt('^'), condition='^', action=_and_then('[.'))
	with META.condition('[.') as start_class:
		start_class.install_rule(expression=txt(']'), action = _class_initial_close_bracket)
		start_class.install_rule(expression=txt('-'), action = _arbitrary_character)
		start_class.install_rule(expression=ref('ANY'), action = _instead('['))
	with META.condition('[') as in_class:
		in_class.install_rule(expression=txt(']'), action=_and_then(None))
		in_class.install_rule(expression=txt('&&'), action=_metatoken)
		in_class.install_rule(expression=txt('&&^'), trail=-1, action=_and_then('^'))
		in_class.install_rule(expression=txt('-'), action=_metatoken)
		in_class.install_rule(expression=txt('-]'), trail=-1, action=_arbitrary_character)
	with META.condition(None, '[') as anywhere:
		anywhere.install_rule(expression=seq(txt('{'), ref('alpha'), regular.Plus(ref('word')), txt('}'), ), action=_bracket_reference)
		whack = txt('\\')
		for c, n in [('x', 2), ('u', 4), ('U', 8)]: META.install_rule(expression=seq(whack, txt(c), regular.Counted(ref('xdigit'), n, n)), action=_hex_escape)
		anywhere.install_rule(expression=seq(whack, txt('c'), regular.CharClass([64, 128])), action=_control)
		anywhere.install_rule(expression=seq(whack, ref('alnum')), action=_shorthand_reference)
		anywhere.install_rule(expression=seq(whack, dot), action=_arbitrary_escape)
		anywhere.install_rule(expression=dot, action=_arbitrary_character)
	with META.condition('{') as brace:
		brace.install_rule(expression=regular.Plus(ref('digit')), action=_number)
		brace.install_rule(expression=txt(','), action=_metatoken)
		brace.install_rule(expression=txt('}'), action=_and_then(None))
	
	PRELOAD['ASCII']['R'] = rex.parse(META.scan(r'\r?\n|\r', env=PRELOAD['ASCII']), language='Regular')
_BEGIN_()



import setuptools

setuptools.setup(
	name='booze-tools',
	version='0.4.1.post1',
	packages=[
		'boozetools',
		'boozetools.macroparse',
		'boozetools.parsing',
		'boozetools.scanning',
		'boozetools.support',
	],
	description='A panoply of tools for parsing and lexical analysis',
	long_description=open('README.md').read(),
	long_description_content_type="text/markdown",
	url="https://github.com/kjosib/booze-tools",
	classifiers=[
		"Programming Language :: Python :: 3.7",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Topic :: Software Development :: Compilers",
		"Development Status :: 4 - Beta",
    ],
)


"""
Parsing literature is rife with different approaches to encode scanner and parser tables
compactly, with all manner of trade-offs between size and access time.

It may be said that today's fast machines and big memories leave little need to compress these
tables. However, compression can help complex tables fit in cache. Alternatively, compact
forms may be used for storage and interchange while implementations reprocess the tables
for quicker operation despite a bigger RAM footprint: this still saves disk space.

Typically, tables used in different contexts have distinct structural characteristics which
make it more or less effective to compress them in particular ways. This module implements a
few typical approaches to compressing scanner and parser tables. The corresponding ways to
use them are implemented in the runtime.py module.

As a nod to runtime efficiency, offsets useful for a displacement table are precomputed here
and provided for later.
"""

import collections, bisect
from typing import List, Optional, Tuple, Sequence, Set
from . import foundation, expansion

VERBOSE = False

###
###  A few convenient abbreviations: Skip them on first reading.
###

def most_common(row, *, default=0):
	""" Return the most-common element in the array; break ties arbitrarily. """
	C = collections.Counter(row)
	return C.most_common(1)[0][0] if C else default

def is_homogeneous(row): return len(row) < 2 or all(r == row[0] for r in row)

def heavy_rows(matrix):
	""" Return the indices of rows that are more-than-half full. Fullness == Truthiness. """
	return [r for r, row in enumerate(matrix) if sum(map(bool, row)) * 2 > len(row)]

def lighten(row):
	""" Preserve `None` (as don't-care); logical-NOT the rest. """
	return tuple(None if x is None else int(not x) for x in row)

def measure_approximate_cost(structure):
	""" Various bits estimate the size of the structures they return. This makes that consistent. """
	if isinstance(structure, (list, tuple)): return 1 + sum(map(measure_approximate_cost, structure))
	elif isinstance(structure, dict): return len(structure) + sum(map(measure_approximate_cost, structure.values()))
	elif isinstance(structure, int) or structure is None: return 1
	else: assert False, type(structure)

###
###  Leitmotifs: Recognizable themes that get used over and over...
###

def find_row_equivalence(matrix, do_not_care) -> Tuple[List[int], List[List]]:
	"""
	:param matrix: A rank-two tensor. Iterable of rows.
	:param do_not_care: Cells with this value may be assigned any convenient value for the purpose.
	"""
	index, classes = [], []
	for row in matrix:
		for class_id, candidate_class in enumerate(classes):
			if all(a == b or do_not_care in (a, b) for a, b in zip(row, candidate_class)):
				index.append(class_id)
				for c, value in enumerate(row):
					if value != do_not_care: candidate_class[c] = value
				break
		else:
			index.append(foundation.allocate(classes, list(row)))
	return index, classes

def find_column_equivalence(matrix, do_not_care) -> Tuple[List[int], List[Sequence]]:
	""" Adapt find_row_equivalence (above) to work on columns instead of rows. """
	index, classes = find_row_equivalence(zip(*matrix), do_not_care)
	return index, list(zip(*classes))

def first_fit_decreasing(indices:list, *, allow_negative:bool) -> Tuple[List[int], int] :
	"""
	(Aho and Ulman [2]) and (Ziegler [7]) advocate this scheme...
	This function finds a set of displacements such that no two index[i][j]+displacement[i] are the same.
	
	This is part of a strategy for compact representation of a sparse matrix
	with fast O(1) read performance. The input to THIS function is a statement
	about which cells in the original matrix contain non-trivial data:
	`col in indices[row]` exactly when `matrix[row][column]` is non-trivial.

	The result can be used to populate a displacement table, which is a simple kind of perfect-hash.
		* A hash is "perfect" when it requires at most one probe to determine
		the presence or absence of a key; in general these must be precomputed.
		* A hash is "minimal" when (in the lingo of hashing) all the buckets are used.
		* The holy grail of "minimal perfect hashing" generally requires a more complex function.
	
	Rows are considered in decreasing order of density, so if Zipf's law applies then the result
	will tend to be packed fairly densely: i.e. minimal or nearly so. This tends to be true of the
	way this particular function gets used in this library.
	
	The size of the necessary resulting vector is also returned: it would have to be determined again
	in every reasonable case otherwise.
	
	A final note: empty input rows are assigned the offset equal to the size of the implied check-vector.
	"""
	used = set()
	def first_fit(row: List[int]) -> int:
		offset = 0 - min(row) if allow_negative else 0
		while any(r + offset in used for r in row): offset += 1
		used.update(r + offset for r in row)
		return offset
	
	displacements = [0] * len(indices) # Allocate space; it will be filled out-of-order.
	populations = list(map(len, indices))
	schedule = iter(foundation.grade(populations, descending=True))
	for i in schedule:
		if indices[i]: displacements[i] = first_fit(indices[i])
		else: break # All remaining should be set == size of implied check vector.
	size = max(used)+1 if used else 0
	for i in schedule: displacements[i] = size
	if VERBOSE: print("First-Fit Decreasing placed %d entries among %d positions."%(len(used), size))
	return displacements, size

def encode_boolean_field_as_offset_check(matrix:Sequence[Sequence[Optional[bool]]]):
	"""
	This is used in both the scanner and parser tables for keeping track of the error plane.
	This function does not worry about equivalence classification.
	It returns a pair of vectors such that:
		check[offset[row]+column] == column
	is equivalent to:
		bool(matrix[row][column])
	The returned structure is smaller when it has fewer `True` entries to encode.
	"""
	ones = [set(c for c,flag in enumerate(row) if flag) for row in matrix]
	offset, size = first_fit_decreasing(ones, allow_negative=True)
	check = [-1] * size
	for r, (base, cs) in enumerate(zip(offset, ones)):
		for c in cs: check[base+c] = r
	return offset, check

def encode_displacement_function(exceptions:list):
	"""
	This builds a typical textbook-style displacement table: a simple kind of perfect-hash,
	which may not be minimal but it's usually decently close.
	:param exceptions: a list of dictionaries; keys are column numbers, values are destination state numbers.
	No attempt is made to coalesce similar or identical rows here:
		(1) It's the caller's responsibility.
		(2) Only the caller knows the best way to perform such a feat.
	There's no guarantee two distinct rows might not get the same offset, so the only
	reliable check-value is the row number. BISON uses the column number instead,
	but it also enforces distinct offsets for all rows.
	"""
	offset, size = first_fit_decreasing(exceptions, allow_negative=True)
	check = [-1]*size
	value = [0]*size
	for row_id, row_dict in enumerate(exceptions):
		row_offset = offset[row_id]
		for column, entry in row_dict.items():
			index = row_offset + column
			assert check[index] == -1
			check[index] = row_id
			value[index] = entry
	return {'offset': offset, 'check': check, 'value':value}

###
###  Major supporting themes: this is where the alleged "deep magic" lives.
###

def compress_dfa_delta_function(matrix:list) -> dict:
	"""
	Per issue #8: A possibly-novel approach to condensing large scanner tables.
	The method is described at https://github.com/kjosib/booze-tools/issues/8
	"""
	# Splitting the matrix into two planes (residue and exceptions):
	zeros, ones, residue, exceptions = [], [], [], []
	for row in matrix:
		common = [k for k,v in collections.Counter(row).most_common(2) if v > 1]
		zeros.append(common[0])
		ones.append(common[-1]) # Because if len(common) == 1, duplicate it.
		look_up = {v:i for i,v in enumerate(common)}
		residue.append([look_up.get(v) for v in row])
		exceptions.append({c:x for c,x in enumerate(row) if x not in common})
	# Compacting the residue plane (a partially-null boolean matrix) by means of equivalence classification:
	column_class, residue = find_column_equivalence(residue, None)
	for i in heavy_rows(residue):
		zeros[i], ones[i] = ones[i], zeros[i]
		residue[i] = lighten(residue[i])
	row_class, residue = find_row_equivalence(residue, None)
	# Converting the final residue to a set-as-displacement-table representation:
	offset, check = encode_boolean_field_as_offset_check(residue)
	# Tie a pretty bow around it:
	return {
		'exceptions': encode_displacement_function(exceptions),
		'bg': {
			# For most states, the most common entry is the error transition. Thus: exceptions to that rule:
			'zero': list(zip(*[(index, value) for index, value in enumerate(zeros) if value != -1])),
			'one': ones,
			'row_class': row_class, 'col_class': column_class,
			'offset': offset, 'check': check,
		},
	}

def compress_scanner(*, initial:dict, matrix:list, final:dict) -> dict:
	height, width = len(matrix), len(matrix[0])
	delta = compress_dfa_delta_function(matrix)
	metric = measure_approximate_cost(delta)
	if VERBOSE: print("DFA matrix was %d rows, %d cols = %d cells; compact form is about %d cells (%0.2f%%)"%(
		height, width, height * width, metric, (100.0*metric/(height*width))
	))
	return {'delta': delta, 'initial': initial, 'final': list(final.keys()), 'rule': list(final.values()),}

def decompose_by_default_reduction(matrix:list, essential_errors:set, recovering_states):
	"""
	The first phase of compacting a parser's "ACTION" table.
	The idea is to separate the matrix into a "default" entry per row
	and a set of exceptions, which can presumably be stored in a lot
	less space.
	
	:param matrix: The initial ACTION matrix is a row for each state;
		the columns correspond to the terminals, and every cell has
		a specific instruction.
	:param essential_errors: (state, terminal_id) pairs which, due to
		non-assoc declarations, must absolutely reflect a syntax error.
		As such, if those states have a default-reduction, then corresponding
		error entries must explicitly appear in the "residue" matrix
	:param recovering_states: These states are REACHED BY the error token.
		To avoid an excess of false restarts, these states may not use
		a default-reduction.
	:return: default-reduction vector and residue matrix where allowed entries have been turned to `None`.
	"""
	reduce = [most_common([a for a in row if a < 0]) for row in matrix] # Default Reduction Table
	for q in recovering_states: reduce[q] = 0
	residue = [[
		value if (q,column) in essential_errors or value not in (0, default) else None
		for column, value in enumerate(row)
	] for q, (row, default) in enumerate(zip(matrix, reduce))]
	return reduce, residue

def decompose_by_edit_distance(action:List[List[int]], focus:List[Set[int]]):
	"""
	This is one area where performance did become a real problem with a simple approach.
	Therefore, I've thought about things and come up with something I can be reasonably
	proud of. All hail the well-considered loop invariant!
	
	The basic idea is:
		1. The outer loop works subject rows in order of _IN_creasing "population".
		2. The inner loop compares against (already-solved) rows in _DE_creasing population.
		3. (Critical) An arithmetic test shows when the inner loop becomes hopeless.
	
	There are a thousand subtle ways to get this wrong. I've tried most of them.
	The trick to getting it right is stating the correct post-condition in terms
	that are easy to verify. Here's the latest try:
	
	The edit-chain from R covers at least the columns in `focus[R]` and matches `action[R]`.
	A useful intermediate datum is which cells are covered by some fallback chain.
	"""
	
	def find_edits(subject, basis, must_cover, basis_covers):
		# If the basis covers it WRONGLY, it must be an edit.
		# If it must be covered and the basis does not, it's an edit.
		# But if the basis covers it correctly, then it's not an edit.
		return [
			c for c in must_cover|basis_covers
			if not (c in basis_covers and basis[c] == subject[c])
		]
	
	def find_best_fallback():
		subject, must_cover = action[current], focus[current]
		fb, best = -1, {c:subject[c] for c in must_cover}
		for size, candidate in reversed(solved):
			if len(best) + size < len(must_cover): break
			golf = find_edits(subject, action[candidate], must_cover, focus[candidate])
			if len(golf) < len(best): fb,best = candidate,golf
		must_cover.update(focus[fb])
		return fb, {c:subject[c] for c in best}
	
	assert len(action) == len(focus)
	fallback = [-1] * len(action)
	edits = [None] * len(action)
	solved = []
	for current in foundation.grade(list(map(len, focus))):
		fallback[current], edits[current] = find_best_fallback()
		if edits[current]: bisect.insort_right(solved, (len(focus[current]), current))
	return fallback, edits

def compress_action_table(action:List[List[int]], nonassoc_errors:set) -> dict:
	"""
	Produce a compact representation of the "ACTION" table for a typical shift-reduce parser.
	:param action: matrix of parse actions: positive numbers are shifts; negative are reductions, zero is error.
	:param nonassoc_errors:
	    set of pairs of (state_id, terminal_id) which must error on account of
	    non-associativity declarations. They are important for two reasons:
	    1. Mentioned states must not become "interactive".
	    2. These error-cells are likely to be isolated, and so good candidates
	       for the reduction-plane to treat as "don't-care": they'll be picked
	       up in the shift-plane anyway.
	:return: a compact structure.
	"""
	height, width = len(action), len(action[0]) # Stats for comparison to the compression method
	
	reduce = compose_reduce_plane(action, nonassoc_errors)
	model = expansion.parser_reduce_function(**reduce)
	focus = [ # A boolean map of where the model gets it wrong.
		set(c for c,i in enumerate(row) if model(r,c) != i)
		for r,row in enumerate(action)
	]
	fallback, edits = decompose_by_edit_distance(action, focus)
	for q,t in nonassoc_errors:
		if fallback[q] == -1 and not edits[q]:
			fallback[q] = -2 # By the way, is this even possible?
	edit_table = encode_displacement_function(edits)
	result = {'reduce': reduce, 'fallback': fallback, 'edits': edit_table,}
	metric = measure_approximate_cost(result)
	if VERBOSE: print("Action matrix was %d * %d = %d; compressed to %d (%0.2f%%)"%(
		height, width, height*width, metric, (100.0*metric)/(height*width)
	))
	return result

def compress_goto_table(goto_table:list) -> dict:
	"""
	Per issue #4: Look alternately for rows or columns with but a single remaining
	significant value. Record this value in a "quotient" list, together with bookkeeping
	data for the rows and columns. Afterwards, we are left with a much smaller residue matrix;
	typical equivalence-class methods may be used upon it.
	
	I didn't make this up. Unfortunately I cannot recall where I saw it in the literature, for
	it was a very long time ago. If anyone knows who invented this, or what article originally
	introduced it, I would very much like to hear from you.
	"""
	
	def remaining(index): return [r for r, x in enumerate(index) if x is None]
	def significant_cells(index, vector): return [v for x,v in zip(index, vector) if x is None and bool(v)]
	def homogenize(vector):
		if len(vector) < 1: return -1
		if is_homogeneous(vector): return vector[0]
	def compact(target_index, second_index, read):
		# This takes some pains to keep the quotient list small. It could be very slightly better...
		tmp = [(homogenize(significant_cells(second_index, read(r))), r) for r in remaining(target_index)]
		for q, i in sorted((q, i) for (q,i) in tmp if q is not None):
			if quotient[-1] != q: quotient.append(q)
			target_index[i] = len(quotient) - 1
	
	# Find rows/columns to zap, accumulating the "quotient" list:
	height, width = len(goto_table), len(goto_table[0])
	row_index, col_index, quotient = [None]*height, [None]*width, [-1]
	while True:
		hi_water_mark = len(quotient)
		compact(row_index, col_index, goto_table.__getitem__)
		compact(col_index, row_index, lambda c:[row[c] for row in goto_table])
		if len(quotient) == hi_water_mark: break
	
	# Capture the much smaller residue matrix:
	row_residue = remaining(row_index)
	column_residue = remaining(col_index)
	residue_matrix = [[row[e] for e in column_residue] for row in [goto_table[e] for e in row_residue]]
	
	# Minimize the residue.
	row_class, minimal_rows = find_row_equivalence(residue_matrix, 0)
	col_class, minimal_columns = find_row_equivalence(zip(*minimal_rows), 0)
	
	# Try to figure the ideal column-class ordering to make it fit better:
	col_class_offset = foundation.collation([vector.count(0) for vector in minimal_columns])
	minimal_rows = [{col_class_offset[c]:x for c, x in enumerate(row) if x} for row in zip(*minimal_columns)]
	
	# Build a single "residue vector" and row-offsets for efficient packing of the residue.
	row_class_offset, size = first_fit_decreasing(minimal_rows, allow_negative=False)
	residue = [0]*size
	for r_off, row in zip(row_class_offset, minimal_rows):
		for c_off, value in row.items():
			if value: residue[r_off+c_off] = value
	
	# Fill the holes in the foo_index vectors using the offsets corresponding to the equivalence classes:
	for state_id, cls_id in zip(row_residue, row_class):
		row_index[state_id] = row_class_offset[cls_id] + hi_water_mark
	for nonterm_id, cls_id,  in zip(column_residue, col_class):
		col_index[nonterm_id] = col_class_offset[cls_id] + hi_water_mark
	
	# Wrap up and return.
	if VERBOSE: print("GOTO table original size: %d rows, %d columns -> %d cells"%(height, width, height * width))
	result = {'row_index': row_index, 'col_index': col_index, 'quotient': quotient+residue, 'mark': hi_water_mark}
	metric = measure_approximate_cost(result)
	if VERBOSE: print("GOTO compact size: %d (%.2f%%)"%(metric, 100.0*metric/(height * width)))
	return result

def compose_reduce_plane(action, nonassoc_errors):
	"""
	What's going on here?
	I want to try to retain the error entries, even in states with a default reduction.
	The glimmer of hope centers on a semantic partitioning of the ACTION table.
	"""
	
	def decide(i, dr):
		if i == 0: return False
		if i == dr: return True
		return None
	
	def best_d_reduce(row):
		x = [a for a in row if a < 0]
		if x:
			i, count = collections.Counter(x).most_common(1)[0]
			if count > 1: return i
		return 0
	
	d_reduce = list(map(best_d_reduce, action))  # Default Reduction Table
	r_plane = [
		[decide(i, dr) for i in row]
		for row, dr in zip(action, d_reduce)
	]
	for q,t in nonassoc_errors: r_plane[q][t] = None # These are likely to be one-offs.
	col_class, minimal_cols = find_row_equivalence(zip(*r_plane), None)
	midpoint = list(zip(*minimal_cols))
	# hot = compaction.heavy_rows(midpoint)
	# for r in hot: midpoint[r] = compaction.lighten(midpoint[r])
	row_class, minimal_rows = find_row_equivalence(midpoint, None)
	offset, check = encode_boolean_field_as_offset_check(minimal_rows)
	return {
		'd_reduce': d_reduce,
		'row_class': row_class,
		'col_class': col_class,
		'offset': offset,
		'check': check,
	}



"""
This module may still be doing too many different things. I'll probably break it up further in the future.

1. It supplies one means to bind rule-actions to driver-objects. (It would be nice to support a few more styles.)
2. It provides a (maybe) convenient runtime interface to the most common use cases.
"""

import inspect, functools, sys

from . import interfaces, expansion, failureprone
from ..scanning import recognition
from ..parsing import shift_reduce


class BoundScanRules(interfaces.ScanRules):
	"""
	This binds symbolic scan-action specifications to a specific "driver" or
	context object. It cannot act alone, but works in concert with the
	CompactDFA (in module support.expansion) and the generic scan algorithm.
	
	Also, it needs to be more like parse_action_bindings, not a class.
	"""
	def __init__(self, *, action:dict, driver:object):
		
		def bind(message, parameter):
			method_name = 'scan_' + message
			if hasattr(driver, method_name):
				fn = getattr(driver, method_name)
			elif default_method is None:
				raise ValueError("IterableScanner driver has neither method %r nor %r." % (method_name, default_method_name))
			else:
				fn = functools.partial(default_method, message)
				method_name = "%s(%r, ...)"%(default_method_name, message)
			arity = len(inspect.signature(fn).parameters)
			if arity == 1:
				if parameter is None: return fn
				else: raise ValueError("Message %r is used with parameter %r but handler %r only takes one argument (the scan state) and needs to take a second (the message parameter)." % (message, parameter, method_name))
			elif arity == 2: return lambda scan_state: fn(scan_state, parameter)
			else: raise ValueError("Scan handler %r takes %d arguments, but needs to take %s." % (method_name, arity, ['two', 'one or two'][parameter is None]))
		
		default_method_name = 'default_scan_action'
		default_method = getattr(driver, default_method_name, None)
		self._trail       = action['trail']
		self._line_number = action['line_number']
		self.__methods = list(map(bind, action['message'], action['parameter']))
		self.get_trailing_context = self._trail.__getitem__
		
	def get_trailing_context(self, rule_id: int):
		""" NB: This gets overwritten by a direct bound-method on the trailing-context list. """
		return self._trail[rule_id]
	
	def invoke(self, scan_state: interfaces.Scanner, rule_id:int):
		return self.__methods[rule_id](scan_state)

class TypicalApplication(interfaces.ScanErrorListener, interfaces.ParseErrorListener):
	"""
	This class aims to provide a simple basis for extension and reasonable
	defaults for a variety of common parsing applications. Those with unusual
	requirements should consider taking this to bits.
	
	THIS MAY SEEM like a method-as-object (anti)pattern. HOWEVER, the real
	point is a whole mess of configuration and cooperation in one place.
	"""
	
	source: failureprone.SourceText
	yy: interfaces.Scanner
	exception: Exception
	
	def __init__(self, tables):
		assert list(tables.get('version', [])) == [0,0,1], 'Data table version mismatch: '+repr(tables.get('version'))
		scanner_tables = tables['scanner']
		self.__dfa = expansion.CompactDFA(dfa=scanner_tables['dfa'], alphabet=scanner_tables['alphabet'])
		self.__rules = BoundScanRules(action=scanner_tables['action'], driver=self)
		self.__hfa = expansion.CompactHandleFindingAutomaton(tables['parser'])
		self.__combine = parse_action_bindings(self, self.__hfa.message_catalog)
	
	def parse(self, text: str, *, line_breaks='normal', filename: str = None, start=None, language=None):
		if start is None: start = interfaces.DEFAULT_INITIAL_CONDITION
		self.source = failureprone.SourceText(text, line_breaks=line_breaks, filename=filename)
		self.yy = recognition.IterableScanner(text=text, automaton=self.__dfa, rules=self.__rules, start=start, on_error=self)
		return shift_reduce.parse(self.__hfa, self.__combine, self.yy, language=language, on_error=self)
	
	def unexpected_token(self, kind, semantic, pds):
		self.source.complain(*self.yy.current_span(), message="Unexpected token %r" % kind)
		# stack_symbols = list(map(self.table.get_breadcrumb, pds.path_from_root()))[1:]
		# self.exception = interfaces.ParseError(stack_symbols, kind, semantic)
		# print("Parsing condition was:\n", self.exception.condition(), file=sys.stderr)
	
	def unexpected_eof(self, pds):
		self.unexpected_token(interfaces.END_OF_TOKENS, None, pds)
	
	def will_recover(self, tokens):
		if len(tokens) >= 3: print("Trying to recover.", file=sys.stderr)
		
	def did_not_recover(self):
		print("Could not recover.", file=sys.stderr)
	
	def rule_exception(self, ex: Exception, message, args):
		self.source.complain(*self.yy.current_span(), message="During " + repr(message))
		raise ex from None
	
	# TODO: By the way, it's no longer clear the scanner should pass `self` as a parameter.
	
	def scan_blocked(self, yy: interfaces.Scanner):
		self.source.complain(yy.current_position(), "Lexical scan got stuck.")
	
	def scan_exception(self, yy:interfaces.Scanner, rule_id:int, ex:Exception):
		self.source.complain(*yy.current_span())
		raise ex from None


def parse_action_bindings(driver, message_catalog):
	"""
	Build a reduction function (combiner) for the parse engine out of an arbitrary Python object.
	Because this checks the message catalog it can be sure the method lookup will never fail.
	"""
	assert isinstance(message_catalog, list), type(message_catalog)
	def bind(message):
		if message is None: return None
		attr = 'action_' + message[1:] if message.startswith(':') else 'parse_' + message
		try: return getattr(driver, attr)
		except AttributeError as ex:
			def fail(*items): raise ex
			return fail
	dispatch = [bind(message) for message in message_catalog]
	def combine(cid:int, args):
		message = dispatch[cid]
		if message is None: return tuple(args) # The null check is like one bytecode and very fast.
		return message(*args)
	return combine

